import _defineProperty from "@babel/runtime/helpers/esm/defineProperty";
import { log, assert } from '@luma.gl/api';
import GL from '@luma.gl/constants';
import { getWebGL2Context, assertWebGL2Context } from '@luma.gl/webgl';
import { getKey } from '../webgl-utils/constants-to-keys';
import Renderbuffer from './renderbuffer';
import { clear, clearBuffer } from './clear';
import { WebGLDevice, WEBGLFramebuffer } from '@luma.gl/webgl';
const ERR_MULTIPLE_RENDERTARGETS = 'Multiple render targets not supported';

function getDefaultProps(props) {
  const newProps = { ...props
  };
  const {
    color = true,
    depth = true,
    stencil = false
  } = props;

  if (props.attachments) {
    newProps.depthStencilAttachment = undefined;
    newProps.colorAttachments = newProps.colorAttachments || [];

    for (const [attachmentPoint, attachment] of Object.entries(props.attachments)) {
      switch (Number(attachmentPoint)) {
        case GL.DEPTH_ATTACHMENT:
        case GL.STENCIL_ATTACHMENT:
        case GL.DEPTH_STENCIL_ATTACHMENT:
          newProps.depthStencilAttachment = attachment;
          break;

        default:
          newProps.colorAttachments.push(attachment);
          break;
      }
    }

    delete newProps.attachments;
    return newProps;
  }

  if (color) {
    newProps.colorAttachments = newProps.colorAttachments || ['rgba8unorm-unsized'];
  }

  if (depth && stencil) {
    newProps.depthStencilAttachment = newProps.depthStencilAttachment || 'depth24plus-stencil8';
  } else if (depth) {
    newProps.depthStencilAttachment = newProps.depthStencilAttachment || 'depth16unorm';
  } else if (stencil) {
    newProps.depthStencilAttachment = newProps.depthStencilAttachment || 'stencil8';
  }

  return newProps;
}

export default class ClassicFramebuffer extends WEBGLFramebuffer {
  static isSupported(gl, options) {
    return isFloatColorBufferSupported(gl, options);
  }

  static getDefaultFramebuffer(gl) {
    const webglDevice = WebGLDevice.attach(gl);
    webglDevice.defaultFramebuffer = webglDevice.defaultFramebuffer || new ClassicFramebuffer(gl, {
      id: 'default-framebuffer',
      handle: null,
      attachments: {},
      check: false
    });
    return webglDevice.defaultFramebuffer;
  }

  get MAX_COLOR_ATTACHMENTS() {
    const gl2 = assertWebGL2Context(this.gl);
    return gl2.getParameter(gl2.MAX_COLOR_ATTACHMENTS);
  }

  get MAX_DRAW_BUFFERS() {
    const gl2 = assertWebGL2Context(this.gl);
    return gl2.getParameter(gl2.MAX_DRAW_BUFFERS);
  }

  constructor(gl, props = {}) {
    super(WebGLDevice.attach(gl), getDefaultProps(props));

    _defineProperty(this, "width", null);

    _defineProperty(this, "height", null);

    _defineProperty(this, "attachments", {});

    _defineProperty(this, "readBuffer", GL.COLOR_ATTACHMENT0);

    _defineProperty(this, "drawBuffers", [GL.COLOR_ATTACHMENT0]);

    _defineProperty(this, "ownResources", []);

    this.initialize(props);
    Object.seal(this);
  }

  get color() {
    return this.colorAttachments[0] || null;
  }

  get texture() {
    return this.colorAttachments[0] || null;
  }

  get depth() {
    return this.depthStencilAttachment || null;
  }

  get stencil() {
    return this.depthStencilAttachment || null;
  }

  initialize(props) {
    let {
      attachments = null
    } = props || {};
    const {
      width = 1,
      height = 1,
      check = true,
      readBuffer = undefined,
      drawBuffers = undefined
    } = props || {};
    assert(width >= 0 && height >= 0, 'Width and height need to be integers');
    this.width = width;
    this.height = height;

    if (attachments) {
      for (const attachment in attachments) {
        const target = attachments[attachment];
        const object = Array.isArray(target) ? target[0] : target;
        object.resize({
          width,
          height
        });
      }
    } else {}

    this.update({
      clearAttachments: true,
      attachments,
      readBuffer,
      drawBuffers
    });

    if (attachments && check) {
      this.checkStatus();
    }
  }

  checkStatus() {
    super._checkStatus();

    return this;
  }

  getStatus() {
    const {
      gl
    } = this;
    const prevHandle = gl.bindFramebuffer(GL.FRAMEBUFFER, this.handle);
    const status = gl.checkFramebufferStatus(GL.FRAMEBUFFER);
    gl.bindFramebuffer(GL.FRAMEBUFFER, prevHandle || null);
    return status;
  }

  bind({
    target = GL.FRAMEBUFFER
  } = {}) {
    this.gl.bindFramebuffer(target, this.handle);
    return this;
  }

  unbind({
    target = GL.FRAMEBUFFER
  } = {}) {
    this.gl.bindFramebuffer(target, null);
    return this;
  }

  _setReadBuffer(readBuffer) {
    const gl2 = getWebGL2Context(this.gl);

    if (gl2) {
      gl2.readBuffer(readBuffer);
    } else {
      assert(readBuffer === GL.COLOR_ATTACHMENT0 || readBuffer === GL.BACK, ERR_MULTIPLE_RENDERTARGETS);
    }
  }

  _setDrawBuffers(drawBuffers) {
    const {
      gl
    } = this;
    const gl2 = assertWebGL2Context(gl);

    if (gl2) {
      gl2.drawBuffers(drawBuffers);
    } else {
      const ext = gl.getExtension('WEBGL_draw_buffers');

      if (ext) {
        ext.drawBuffersWEBGL(drawBuffers);
      } else {
        assert(drawBuffers.length === 1 && (drawBuffers[0] === GL.COLOR_ATTACHMENT0 || drawBuffers[0] === GL.BACK), ERR_MULTIPLE_RENDERTARGETS);
      }
    }
  }

  _createHandle() {
    return this.gl.createFramebuffer();
  }

  _deleteHandle() {
    this.gl.deleteFramebuffer(this.handle);
  }

  _bindHandle(handle) {
    return this.gl.bindFramebuffer(GL.FRAMEBUFFER, handle);
  }

  update(options) {
    const {
      attachments = {},
      readBuffer,
      drawBuffers,
      clearAttachments = false,
      resizeAttachments = true
    } = options;
    this.attach(attachments, {
      clearAttachments,
      resizeAttachments
    });
    const {
      gl
    } = this;
    const prevHandle = gl.bindFramebuffer(GL.FRAMEBUFFER, this.handle);

    if (readBuffer) {
      this._setReadBuffer(readBuffer);

      this.readBuffer = readBuffer;
    }

    if (drawBuffers) {
      this._setDrawBuffers(drawBuffers);

      this.drawBuffers = drawBuffers;
    }

    gl.bindFramebuffer(GL.FRAMEBUFFER, prevHandle || null);
    return this;
  }

  attach(attachments, {
    clearAttachments = false,
    resizeAttachments = true
  } = {}) {
    const newAttachments = {};

    if (clearAttachments) {
      Object.keys(this.attachments).forEach(key => {
        newAttachments[key] = null;
      });
    }

    Object.assign(newAttachments, attachments);
    const prevHandle = this.gl.bindFramebuffer(GL.FRAMEBUFFER, this.handle);

    for (const key in newAttachments) {
      assert(key !== undefined, 'Misspelled framebuffer binding point?');
      const attachment = Number(key);
      const descriptor = newAttachments[attachment];
      let object = descriptor;

      if (!object) {
        this._unattach(attachment);
      } else {
        object = this._attachOne(attachment, object);
        this.attachments[attachment] = object;
      }

      if (resizeAttachments && object) {
        object.resize({
          width: this.width,
          height: this.height
        });
      }
    }

    this.gl.bindFramebuffer(GL.FRAMEBUFFER, prevHandle || null);
    Object.assign(this.attachments, attachments);
    Object.keys(this.attachments).filter(key => !this.attachments[key]).forEach(key => {
      delete this.attachments[key];
    });
  }

  clear(options) {
    const {
      color,
      depth,
      stencil,
      drawBuffers = []
    } = options;
    const prevHandle = this.gl.bindFramebuffer(GL.FRAMEBUFFER, this.handle);

    if (color || depth || stencil) {
      clear(this.gl, {
        color,
        depth,
        stencil
      });
    }

    drawBuffers.forEach((value, drawBuffer) => {
      clearBuffer(this.gl, {
        drawBuffer,
        value
      });
    });
    this.gl.bindFramebuffer(GL.FRAMEBUFFER, prevHandle || null);
    return this;
  }

  blit(opts = {}) {
    log.error('Framebuffer.blit({...}) is no logner supported, use blit(source, target, opts)')();
    return null;
  }

  invalidate(options) {
    const {
      attachments = [],
      x = 0,
      y = 0,
      width,
      height
    } = options;
    const gl2 = assertWebGL2Context(this.gl);
    const prevHandle = gl2.bindFramebuffer(GL.READ_FRAMEBUFFER, this.handle);
    const invalidateAll = x === 0 && y === 0 && width === undefined && height === undefined;

    if (invalidateAll) {
      gl2.invalidateFramebuffer(GL.READ_FRAMEBUFFER, attachments);
    } else {
      gl2.invalidateFramebuffer(GL.READ_FRAMEBUFFER, attachments, x, y, width, height);
    }

    gl2.bindFramebuffer(GL.READ_FRAMEBUFFER, prevHandle);
    return this;
  }

  getAttachmentParameter(attachment, pname, keys) {
    let value = this._getAttachmentParameterFallback(pname);

    if (value === null) {
      this.gl.bindFramebuffer(GL.FRAMEBUFFER, this.handle);
      value = this.gl.getFramebufferAttachmentParameter(GL.FRAMEBUFFER, attachment, pname);
      this.gl.bindFramebuffer(GL.FRAMEBUFFER, null);
    }

    if (keys && value > 1000) {
      value = getKey(this.gl, value);
    }

    return value;
  }

  getAttachmentParameters(attachment = GL.COLOR_ATTACHMENT0, keys, parameters = this.constructor.ATTACHMENT_PARAMETERS || []) {
    const values = {};

    for (const pname of parameters) {
      const key = keys ? getKey(this.gl, pname) : pname;
      values[key] = this.getAttachmentParameter(attachment, pname, keys);
    }

    return values;
  }

  getParameters(keys = true) {
    const attachments = Object.keys(this.attachments);
    const parameters = {};

    for (const attachmentName of attachments) {
      const attachment = Number(attachmentName);
      const key = keys ? getKey(this.gl, attachment) : attachment;
      parameters[key] = this.getAttachmentParameters(attachment, keys);
    }

    return parameters;
  }

  show() {
    if (typeof window !== 'undefined') {}

    return this;
  }

  log(logLevel = 0, message = '') {
    if (logLevel > log.level || typeof window === 'undefined') {
      return this;
    }

    message = message || "Framebuffer ".concat(this.id);
    return this;
  }

  _unattach(attachment) {
    const oldAttachment = this.attachments[attachment];

    if (!oldAttachment) {
      return;
    }

    if (oldAttachment instanceof Renderbuffer) {
      this.gl.framebufferRenderbuffer(GL.FRAMEBUFFER, attachment, GL.RENDERBUFFER, null);
    } else {
      this.gl.framebufferTexture2D(GL.FRAMEBUFFER, attachment, GL.TEXTURE_2D, null, 0);
    }

    delete this.attachments[attachment];
  }

  _getAttachmentParameterFallback(pname) {
    const webglDevice = WebGLDevice.attach(this.gl);
    const features = webglDevice.features;

    switch (pname) {
      case GL.FRAMEBUFFER_ATTACHMENT_TEXTURE_LAYER:
        return !features.has('webgl2') ? 0 : null;

      case GL.FRAMEBUFFER_ATTACHMENT_RED_SIZE:
      case GL.FRAMEBUFFER_ATTACHMENT_GREEN_SIZE:
      case GL.FRAMEBUFFER_ATTACHMENT_BLUE_SIZE:
      case GL.FRAMEBUFFER_ATTACHMENT_ALPHA_SIZE:
      case GL.FRAMEBUFFER_ATTACHMENT_DEPTH_SIZE:
      case GL.FRAMEBUFFER_ATTACHMENT_STENCIL_SIZE:
        return !features.has('webgl2') ? 8 : null;

      case GL.FRAMEBUFFER_ATTACHMENT_COMPONENT_TYPE:
        return !features.has('webgl2') ? GL.UNSIGNED_INT : null;

      case GL.FRAMEBUFFER_ATTACHMENT_COLOR_ENCODING:
        return !features.has('webgl2') && !features.has('texture-formats-srgb-webgl1') ? GL.LINEAR : null;

      default:
        return null;
    }
  }

}

_defineProperty(ClassicFramebuffer, "FRAMEBUFFER_ATTACHMENT_PARAMETERS", [GL.FRAMEBUFFER_ATTACHMENT_OBJECT_NAME, GL.FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE, GL.FRAMEBUFFER_ATTACHMENT_COLOR_ENCODING, GL.FRAMEBUFFER_ATTACHMENT_RED_SIZE, GL.FRAMEBUFFER_ATTACHMENT_GREEN_SIZE, GL.FRAMEBUFFER_ATTACHMENT_BLUE_SIZE, GL.FRAMEBUFFER_ATTACHMENT_ALPHA_SIZE, GL.FRAMEBUFFER_ATTACHMENT_DEPTH_SIZE, GL.FRAMEBUFFER_ATTACHMENT_STENCIL_SIZE]);

function isFloatColorBufferSupported(gl, options) {
  const {
    colorBufferFloat,
    colorBufferHalfFloat
  } = options;
  let supported = true;

  if (colorBufferFloat) {
    supported = Boolean(gl.getExtension('EXT_color_buffer_float') || gl.getExtension('WEBGL_color_buffer_float') || gl.getExtension('OES_texture_float'));
  }

  if (colorBufferHalfFloat) {
    supported = supported && Boolean(gl.getExtension('EXT_color_buffer_float') || gl.getExtension('EXT_color_buffer_half_float'));
  }

  return supported;
}
//# sourceMappingURL=framebuffer.js.map