import _defineProperty from "@babel/runtime/helpers/esm/defineProperty";

let _Symbol$toStringTag;

import { log, assert, uid, cast } from '@luma.gl/api';
import GL from '@luma.gl/constants';
import { parseUniformName, getUniformSetter } from './uniforms';
import ProgramConfiguration from './program-configuration';
import { copyUniform, checkUniformValues } from './uniforms';
import { getKey } from '../webgl-utils/constants-to-keys';
import { getPrimitiveDrawMode } from '../webgl-utils/attribute-utils';
import VertexArray from './vertex-array';
import { isWebGL2, assertWebGL2Context } from '@luma.gl/webgl';
import { withParameters, withDeviceParameters } from '@luma.gl/webgl';
import { WebGLDevice, WEBGLRenderPipeline, WEBGLTexture, WEBGLFramebuffer } from '@luma.gl/webgl';
const LOG_PROGRAM_PERF_PRIORITY = 4;
const GL_SEPARATE_ATTRIBS = 0x8c8d;

function getRenderPipelineProps(device, props) {
  const newProps = { ...props
  };

  if (typeof props.vs === 'string') {
    newProps.vs = device.createShader({
      id: "".concat(props.id, "-vs"),
      source: props.vs,
      stage: 'vertex'
    });
  }

  if (typeof props.fs === 'string') {
    newProps.fs = device.createShader({
      id: "".concat(props.id, "-fs"),
      source: props.fs,
      stage: 'fragment'
    });
  }

  return newProps;
}

_Symbol$toStringTag = Symbol.toStringTag;
export default class Program extends WEBGLRenderPipeline {
  get [_Symbol$toStringTag]() {
    return 'Program';
  }

  constructor(device, props) {
    super(WebGLDevice.attach(device), getRenderPipelineProps(WebGLDevice.attach(device), props));

    _defineProperty(this, "gl", void 0);

    _defineProperty(this, "gl2", void 0);

    _defineProperty(this, "configuration", void 0);

    _defineProperty(this, "hash", void 0);

    _defineProperty(this, "uniforms", void 0);

    _defineProperty(this, "varyings", void 0);

    _defineProperty(this, "_textureUniforms", void 0);

    _defineProperty(this, "_isCached", false);

    _defineProperty(this, "_textureIndexCounter", 0);

    _defineProperty(this, "_uniformCount", 0);

    _defineProperty(this, "_uniformSetters", void 0);

    _defineProperty(this, "_parameters", void 0);

    _defineProperty(this, "vertexArray", void 0);

    this.gl = this.device.gl;
    this.gl2 = this.device.gl2;
    this._parameters = props.parameters;
    this.vertexArray = new VertexArray(this.device.gl);
    this.initialize(props);
    Object.seal(this);

    this._setId(props.id);
  }

  destroy(options = {}) {
    if (!this._isCached) {
      super.destroy();
    }
  }

  initialize(props) {
    props = getRenderPipelineProps(this.device, props);
    const {
      hash,
      vs,
      fs,
      varyings,
      bufferMode = GL_SEPARATE_ATTRIBS
    } = props;
    this.vs = cast(vs);
    this.fs = fs && cast(fs);
    this.hash = hash || '';
    assert(this.vs.stage === 'vertex');
    this.uniforms = {};
    this._textureUniforms = {};

    if (varyings && varyings.length > 0) {
      assertWebGL2Context(this.gl);
      this.varyings = varyings;
      this.gl2.transformFeedbackVaryings(this.handle, varyings, bufferMode);
    }

    this._readUniformLocationsFromLinkedProgram();

    this.configuration = new ProgramConfiguration(this);
    return this.setProps(props);
  }

  setProps(props) {
    if ('uniforms' in props) {
      this.setUniforms(props.uniforms);
    }

    return this;
  }

  setAttributes(attributes) {
    this.vertexArray.setAttributes(attributes);
  }

  draw(options) {
    const {
      logPriority,
      drawMode = GL.TRIANGLES,
      vertexCount,
      offset = 0,
      start,
      end,
      isIndexed = false,
      indexType = GL.UNSIGNED_SHORT,
      instanceCount = 0,
      isInstanced = instanceCount > 0,
      vertexArray = null,
      transformFeedback,
      framebuffer,
      uniforms,
      samplers
    } = options;
    let {
      parameters = {}
    } = options;

    if (uniforms || samplers) {
      log.deprecated('Program.draw({uniforms})', 'Program.setUniforms(uniforms)')();
      this.setUniforms(uniforms || {});
    }

    if (log.priority >= logPriority) {
      const fb = framebuffer ? framebuffer.id : 'default';
      const message = "mode=".concat(getKey(this.gl, drawMode), " verts=").concat(vertexCount, " ") + "instances=".concat(instanceCount, " indexType=").concat(getKey(this.gl, indexType), " ") + "isInstanced=".concat(isInstanced, " isIndexed=").concat(isIndexed, " ") + "Framebuffer=".concat(fb);
      log.log(logPriority, message)();
    }

    assert(vertexArray);
    this.gl.useProgram(this.handle);

    if (!this._areTexturesRenderable() || vertexCount === 0 || isInstanced && instanceCount === 0) {
      return false;
    }

    vertexArray.bindForDraw(vertexCount, instanceCount, () => {
      if (framebuffer !== undefined) {
        parameters = Object.assign({}, parameters, {
          framebuffer
        });
      }

      if (transformFeedback) {
        const primitiveMode = getPrimitiveDrawMode(drawMode);
        transformFeedback.begin(primitiveMode);
      }

      this._bindTextures();

      withDeviceParameters(this.device, this._parameters, () => {
        withParameters(this.gl, parameters, () => {
          if (isIndexed && isInstanced) {
            this.gl2.drawElementsInstanced(drawMode, vertexCount, indexType, offset, instanceCount);
          } else if (isIndexed && isWebGL2(this.gl) && !isNaN(start) && !isNaN(end)) {
            this.gl2.drawRangeElements(drawMode, start, end, vertexCount, indexType, offset);
          } else if (isIndexed) {
            this.gl.drawElements(drawMode, vertexCount, indexType, offset);
          } else if (isInstanced) {
            this.gl2.drawArraysInstanced(drawMode, offset, vertexCount, instanceCount);
          } else {
            this.gl.drawArrays(drawMode, offset, vertexCount);
          }
        });
      });

      if (transformFeedback) {
        transformFeedback.end();
      }
    });
    return true;
  }

  setUniforms(uniforms = {}) {
    if (log.priority >= 2) {
      checkUniformValues(uniforms, this.id, this._uniformSetters);
    }

    this.gl.useProgram(this.handle);

    for (const uniformName in uniforms) {
      const uniform = uniforms[uniformName];
      const uniformSetter = this._uniformSetters[uniformName];

      if (uniformSetter) {
        let value = uniform;
        let textureUpdate = false;

        if (value instanceof WEBGLFramebuffer) {
          value = value.texture;
        }

        if (value instanceof WEBGLTexture) {
          textureUpdate = this.uniforms[uniformName] !== uniform;

          if (textureUpdate) {
            if (uniformSetter.textureIndex === undefined) {
              uniformSetter.textureIndex = this._textureIndexCounter++;
            }

            const texture = value;
            const {
              textureIndex
            } = uniformSetter;
            texture.bind(textureIndex);
            value = textureIndex;
            this._textureUniforms[uniformName] = texture;
          } else {
            value = uniformSetter.textureIndex;
          }
        } else if (this._textureUniforms[uniformName]) {
          delete this._textureUniforms[uniformName];
        }

        if (uniformSetter(value) || textureUpdate) {
          copyUniform(this.uniforms, uniformName, uniform);
        }
      }
    }

    return this;
  }

  _areTexturesRenderable() {
    let texturesRenderable = true;

    for (const uniformName in this._textureUniforms) {
      const texture = this._textureUniforms[uniformName];
      texture.update();
      texturesRenderable = texturesRenderable && texture.loaded;
    }

    return texturesRenderable;
  }

  _bindTextures() {
    for (const uniformName in this._textureUniforms) {
      const textureIndex = this._uniformSetters[uniformName].textureIndex;

      this._textureUniforms[uniformName].bind(textureIndex);
    }
  }

  _setId(id) {
    if (!id) {
      const programName = this._getName();

      this.id = uid(programName);
    }
  }

  _getName() {
    let programName = this.vs.id || this.fs.id;
    programName = programName.replace(/shader/i, '');
    programName = programName ? "".concat(programName, "-program") : 'program';
    return programName;
  }

  _getOptionsFromHandle(handle) {
    const shaderHandles = this.gl.getAttachedShaders(handle);
    const opts = {};

    for (const shaderHandle of shaderHandles) {
      const type = this.gl.getShaderParameter(this.handle, GL.SHADER_TYPE);

      switch (type) {
        case GL.VERTEX_SHADER:
          opts.vs = new VertexShader({
            handle: shaderHandle
          });
          break;

        case GL.FRAGMENT_SHADER:
          opts.fs = new FragmentShader({
            handle: shaderHandle
          });
          break;

        default:
      }
    }

    return opts;
  }

  _getParameter(pname) {
    return this.gl.getProgramParameter(this.handle, pname);
  }

  _readUniformLocationsFromLinkedProgram() {
    const {
      gl
    } = this;
    this._uniformSetters = {};
    this._uniformCount = this._getParameter(GL.ACTIVE_UNIFORMS);

    for (let i = 0; i < this._uniformCount; i++) {
      const info = this.gl.getActiveUniform(this.handle, i);
      const {
        name
      } = parseUniformName(info.name);
      let location = gl.getUniformLocation(this.handle, name);
      this._uniformSetters[name] = getUniformSetter(gl, location, info);

      if (info.size > 1) {
        for (let l = 0; l < info.size; l++) {
          location = gl.getUniformLocation(this.handle, "".concat(name, "[").concat(l, "]"));
          this._uniformSetters["".concat(name, "[").concat(l, "]")] = getUniformSetter(gl, location, info);
        }
      }
    }

    this._textureIndexCounter = 0;
  }

  getActiveUniforms(uniformIndices, pname) {
    return this.gl2.getActiveUniforms(this.handle, uniformIndices, pname);
  }

  getUniformBlockIndex(blockName) {
    return this.gl2.getUniformBlockIndex(this.handle, blockName);
  }

  getActiveUniformBlockParameter(blockIndex, pname) {
    return this.gl2.getActiveUniformBlockParameter(this.handle, blockIndex, pname);
  }

  uniformBlockBinding(blockIndex, blockBinding) {
    this.gl2.uniformBlockBinding(this.handle, blockIndex, blockBinding);
  }

}
//# sourceMappingURL=program.js.map