import _defineProperty from "@babel/runtime/helpers/esm/defineProperty";
import { luma, log, requestAnimationFrame, cancelAnimationFrame } from '@luma.gl/api';
import { isWebGL, resetParameters, Query, Framebuffer } from '@luma.gl/gltools';
import { isBrowser } from '@probe.gl/env';
const isPage = isBrowser() && typeof document !== 'undefined';

function getHTMLCanvasElement(canvas) {
  return typeof HTMLCanvasElement !== 'undefined' && canvas instanceof HTMLCanvasElement ? canvas : null;
}

let statIdCounter = 0;
const DEFAULT_CLASSIC_ANIMATION_LOOP_PROPS = {
  onCreateDevice: props => luma.createDevice(props),
  onCreateContext: null,
  onAddHTML: null,
  onInitialize: () => ({}),
  onRender: () => {},
  onFinalize: () => {},
  onError: error => console.error(error),
  device: null,
  useDevicePixels: true,
  autoResizeViewport: true,
  autoResizeDrawingBuffer: true,
  stats: luma.stats.get("animation-loop-".concat(statIdCounter++)),
  gl: null,
  glOptions: {},
  createFramebuffer: false
};
export default class ClassicAnimationLoop {
  constructor(props = {}) {
    _defineProperty(this, "device", void 0);

    _defineProperty(this, "canvas", void 0);

    _defineProperty(this, "props", void 0);

    _defineProperty(this, "animationProps", void 0);

    _defineProperty(this, "framebuffer", null);

    _defineProperty(this, "timeline", null);

    _defineProperty(this, "stats", void 0);

    _defineProperty(this, "cpuTime", void 0);

    _defineProperty(this, "gpuTime", void 0);

    _defineProperty(this, "frameRate", void 0);

    _defineProperty(this, "display", void 0);

    _defineProperty(this, "needsRedraw", 'initialized');

    _defineProperty(this, "_initialized", false);

    _defineProperty(this, "_running", false);

    _defineProperty(this, "_animationFrameId", null);

    _defineProperty(this, "_pageLoadPromise", null);

    _defineProperty(this, "_nextFramePromise", null);

    _defineProperty(this, "_resolveNextFrame", null);

    _defineProperty(this, "_cpuStartTime", 0);

    _defineProperty(this, "_gpuTimeQuery", null);

    _defineProperty(this, "gl", void 0);

    this.props = { ...DEFAULT_CLASSIC_ANIMATION_LOOP_PROPS,
      ...props
    };
    props = this.props;
    let {
      useDevicePixels = true
    } = this.props;

    if ('useDevicePixelRatio' in props) {
      log.deprecated('useDevicePixelRatio', 'useDevicePixels')();
      useDevicePixels = props.useDevicePixelRatio;
    }

    this.device = props.device;
    this.gl = this.device && this.device.gl || props.gl;
    this.stats = props.stats;
    this.cpuTime = this.stats.get('CPU Time');
    this.gpuTime = this.stats.get('GPU Time');
    this.frameRate = this.stats.get('Frame Rate');
    this.setProps({
      autoResizeViewport: props.autoResizeViewport,
      autoResizeDrawingBuffer: props.autoResizeDrawingBuffer,
      useDevicePixels
    });
    this.start = this.start.bind(this);
    this.stop = this.stop.bind(this);
    this._onMousemove = this._onMousemove.bind(this);
    this._onMouseleave = this._onMouseleave.bind(this);
  }

  destroy() {
    this.stop();

    this._setDisplay(null);
  }

  delete() {
    this.destroy();
  }

  setNeedsRedraw(reason) {
    this.needsRedraw = this.needsRedraw || reason;
    return this;
  }

  setProps(props) {
    if ('autoResizeViewport' in props) {
      this.props.autoResizeViewport = props.autoResizeViewport;
    }

    if ('autoResizeDrawingBuffer' in props) {
      this.props.autoResizeDrawingBuffer = props.autoResizeDrawingBuffer;
    }

    if ('useDevicePixels' in props) {
      this.props.useDevicePixels = props.useDevicePixels;
    }

    return this;
  }

  start(opts = {}) {
    this._start(opts);

    return this;
  }

  async _start(props) {
    if (this._running) {
      return this;
    }

    this._running = true;

    try {
      await this._getPageLoadPromise();

      if (!this._running) {
        return null;
      }

      let appContext;

      if (!this._initialized) {
        this._initialized = true;
        await this._createDevice(props);

        this._initialize(props);

        appContext = await this.onInitialize(this.animationProps);

        this._addCallbackData(appContext || {});
      }

      if (!this._running) {
        return null;
      }

      if (appContext !== false) {
        this._cancelAnimationFrame();

        this._requestAnimationFrame();
      }

      return this;
    } catch (error) {
      this.props.onError(error);
      return null;
    }
  }

  redraw() {
    if (this.isContextLost()) {
      return this;
    }

    this._beginTimers();

    this._setupFrame();

    this._updateCallbackData();

    this._renderFrame(this.animationProps);

    this._clearNeedsRedraw();

    if (this._resolveNextFrame) {
      this._resolveNextFrame(this);

      this._nextFramePromise = null;
      this._resolveNextFrame = null;
    }

    this._endTimers();

    return this;
  }

  stop() {
    if (this._running) {
      this._finalizeCallbackData();

      this._cancelAnimationFrame();

      this._nextFramePromise = null;
      this._resolveNextFrame = null;
      this._running = false;
    }

    return this;
  }

  attachTimeline(timeline) {
    this.timeline = timeline;
    return this.timeline;
  }

  detachTimeline() {
    this.timeline = null;
  }

  waitForRender() {
    this.setNeedsRedraw('waitForRender');

    if (!this._nextFramePromise) {
      this._nextFramePromise = new Promise(resolve => {
        this._resolveNextFrame = resolve;
      });
    }

    return this._nextFramePromise;
  }

  async toDataURL() {
    var _getHTMLCanvasElement;

    this.setNeedsRedraw('toDataURL');
    await this.waitForRender();
    return (_getHTMLCanvasElement = getHTMLCanvasElement(this.gl.canvas)) === null || _getHTMLCanvasElement === void 0 ? void 0 : _getHTMLCanvasElement.toDataURL();
  }

  isContextLost() {
    return this.gl.isContextLost();
  }

  onCreateDevice(deviceProps) {
    return this.props.onCreateDevice(deviceProps);
  }

  onInitialize(animationProps) {
    return this.props.onInitialize(animationProps);
  }

  onRender(animationProps) {
    return this.props.onRender(animationProps);
  }

  onFinalize(animationProps) {
    return this.props.onFinalize(animationProps);
  }

  onCreateContext(props) {
    return this.props.onCreateContext(props);
  }

  getHTMLControlValue(id, defaultValue = 1) {
    const element = document.getElementById(id);
    return element ? Number(element.value) : defaultValue;
  }

  _initialize(props) {
    this._createFramebuffer();

    this._startEventHandling();

    this._initializeCallbackData();

    this._updateCallbackData();

    this._resizeCanvasDrawingBuffer();

    this._resizeViewport();

    this._gpuTimeQuery = Query.isSupported(this.gl, ['timers']) ? new Query(this.gl) : null;
  }

  _getPageLoadPromise() {
    if (!this._pageLoadPromise) {
      this._pageLoadPromise = isPage ? new Promise((resolve, reject) => {
        if (isPage && document.readyState === 'complete') {
          resolve(document);
          return;
        }

        window.addEventListener('load', () => {
          resolve(document);
        });
      }) : Promise.resolve({});
    }

    return this._pageLoadPromise;
  }

  _setDisplay(display) {
    if (this.display) {
      this.display.delete();
      this.display.animationLoop = null;
    }

    if (display) {
      display.animationLoop = this;
    }

    this.display = display;
  }

  _requestAnimationFrame() {
    if (!this._running) {
      return;
    }

    this._animationFrameId = requestAnimationFrame(this._animationFrame.bind(this));
  }

  _cancelAnimationFrame() {
    if (this._animationFrameId !== null) {
      return;
    }

    cancelAnimationFrame(this._animationFrameId);
    this._animationFrameId = null;
  }

  _animationFrame() {
    if (!this._running) {
      return;
    }

    this.redraw();

    this._requestAnimationFrame();
  }

  _renderFrame(props) {
    if (this.display) {
      this.display._renderFrame(props);

      return;
    }

    this.onRender(props);
  }

  _clearNeedsRedraw() {
    this.needsRedraw = null;
  }

  _setupFrame() {
    this._resizeCanvasDrawingBuffer();

    this._resizeViewport();

    this._resizeFramebuffer();
  }

  _initializeCallbackData() {
    this.animationProps = {
      device: this.device,
      gl: this.gl,
      stop: this.stop,
      canvas: this.gl.canvas,
      useDevicePixels: this.props.useDevicePixels,
      needsRedraw: null,
      startTime: Date.now(),
      engineTime: 0,
      tick: 0,
      tock: 0,
      timeline: this.timeline,
      animationLoop: this,
      time: 0,
      _mousePosition: null,
      framebuffer: this.framebuffer,
      _timeline: this.timeline,
      _loop: this,
      _animationLoop: this
    };
  }

  _updateCallbackData() {
    const {
      width,
      height,
      aspect
    } = this._getSizeAndAspect();

    if (width !== this.animationProps.width || height !== this.animationProps.height) {
      this.setNeedsRedraw('drawing buffer resized');
    }

    if (aspect !== this.animationProps.aspect) {
      this.setNeedsRedraw('drawing buffer aspect changed');
    }

    this.animationProps.width = width;
    this.animationProps.height = height;
    this.animationProps.aspect = aspect;
    this.animationProps.needsRedraw = this.needsRedraw;
    this.animationProps.engineTime = Date.now() - this.animationProps.startTime;

    if (this.timeline) {
      this.timeline.update(this.animationProps.engineTime);
    }

    this.animationProps.tick = Math.floor(this.animationProps.time / 1000 * 60);
    this.animationProps.tock++;
    this.animationProps.time = this.timeline ? this.timeline.getTime() : this.animationProps.engineTime;
  }

  _finalizeCallbackData() {
    this.onFinalize(this.animationProps);
  }

  _addCallbackData(appContext) {
    if (typeof appContext === 'object' && appContext !== null) {
      this.animationProps = Object.assign({}, this.animationProps, appContext);
    }
  }

  async _createDevice(props) {
    const deviceProps = { ...this.props,
      ...props,
      ...this.props.glOptions
    };
    this.device = await this.onCreateDevice(deviceProps);
    this.gl = this.device.gl;

    if (!isWebGL(this.gl)) {
      throw new Error('ClassicAnimationLoop.onCreateContext - illegal context returned');
    }

    resetParameters(this.gl);

    this._createInfoDiv();
  }

  _createInfoDiv() {
    const canvas = getHTMLCanvasElement(this.gl.canvas);

    if (canvas && this.props.onAddHTML) {
      const wrapperDiv = document.createElement('div');
      document.body.appendChild(wrapperDiv);
      wrapperDiv.style.position = 'relative';
      const div = document.createElement('div');
      div.style.position = 'absolute';
      div.style.left = '10px';
      div.style.bottom = '10px';
      div.style.width = '300px';
      div.style.background = 'white';

      if (canvas) {
        wrapperDiv.appendChild(canvas);
      }

      wrapperDiv.appendChild(div);
      const html = this.props.onAddHTML(div);

      if (html) {
        div.innerHTML = html;
      }
    }
  }

  _getSizeAndAspect() {
    const width = this.gl.drawingBufferWidth;
    const height = this.gl.drawingBufferHeight;
    let aspect = 1;
    const canvas = getHTMLCanvasElement(this.gl.canvas);

    if (canvas && canvas.clientHeight) {
      aspect = canvas.clientWidth / canvas.clientHeight;
    } else if (width > 0 && height > 0) {
      aspect = width / height;
    }

    return {
      width,
      height,
      aspect
    };
  }

  _resizeViewport() {
    if (this.props.autoResizeViewport) {
      this.gl.viewport(0, 0, this.gl.drawingBufferWidth, this.gl.drawingBufferHeight);
    }
  }

  _resizeCanvasDrawingBuffer() {
    if (this.props.autoResizeDrawingBuffer) {
      this.device.canvasContext.resize({
        useDevicePixels: this.props.useDevicePixels
      });
    }
  }

  _beginTimers() {
    this.frameRate.timeEnd();
    this.frameRate.timeStart();

    if (this._gpuTimeQuery && this._gpuTimeQuery.isResultAvailable() && !this._gpuTimeQuery.isTimerDisjoint()) {
      this.stats.get('GPU Time').addTime(this._gpuTimeQuery.getTimerMilliseconds());
    }

    if (this._gpuTimeQuery) {
      this._gpuTimeQuery.beginTimeElapsedQuery();
    }

    this.cpuTime.timeStart();
  }

  _endTimers() {
    this.cpuTime.timeEnd();

    if (this._gpuTimeQuery) {
      this._gpuTimeQuery.end();
    }
  }

  _startEventHandling() {
    const {
      canvas
    } = this.gl;

    if (canvas) {
      canvas.addEventListener('mousemove', this._onMousemove);
      canvas.addEventListener('mouseleave', this._onMouseleave);
    }
  }

  _onMousemove(e) {
    this.animationProps._mousePosition = [e.offsetX, e.offsetY];
  }

  _onMouseleave(e) {
    this.animationProps._mousePosition = null;
  }

  _createFramebuffer() {
    if (this.props.createFramebuffer) {
      this.framebuffer = new Framebuffer(this.gl);
    }
  }

  _resizeFramebuffer() {
    if (this.framebuffer) {
      this.framebuffer.resize({
        width: this.gl.drawingBufferWidth,
        height: this.gl.drawingBufferHeight
      });
    }
  }

}
//# sourceMappingURL=classic-animation-loop.js.map