import { Device } from '@luma.gl/api';
import { WebGLDevice, GLParameters } from '@luma.gl/webgl';
import type Buffer from '../classic/buffer';
import type Framebuffer from '../classic/framebuffer';
import { default as Texture2D } from '../classic/texture-2d';
import { default as TransformFeedback } from '../classic/transform-feedback';
import Model from '../engine/classic-model';
import BufferTransform from './buffer-transform';
import TextureTransform from './texture-transform';
/** Properties for creating Transforms */
export declare type TransformProps = {
    id?: string;
    vs?: string;
    elementCount?: number;
    sourceBuffers?: Record<string, Buffer>;
    feedbackBuffers?: Record<string, string | Buffer | {
        buffer: Buffer;
        byteOffset?: number;
    }>;
    varyings?: string[];
    feedbackMap?: Record<string, string>;
    modules?: object[];
    attributes?: Record<string, any>;
    uniforms?: Record<string, any>;
    defines?: Record<string, any>;
    parameters?: GLParameters;
    discard?: boolean;
    isIndexed?: boolean;
    inject?: Record<string, string>;
    drawMode?: number;
    framebuffer?: Framebuffer;
    _sourceTextures?: Record<string, Texture2D>;
    _targetTexture?: string | Texture2D;
    _targetTextureVarying?: string;
    _swapTexture?: string | null;
    _fs?: string;
    fs?: string;
};
/** Options that can be provided when running a Transform */
export declare type TransformRunOptions = {
    framebuffer?: Framebuffer;
    clearRenderTarget?: boolean;
    uniforms?: Record<string, any>;
    parameters?: Record<string, any>;
    discard?: boolean;
};
/** Options that control drawing a Transform. Used by subclasses to return draw parameters */
export declare type TransformDrawOptions = {
    attributes?: Record<string, any>;
    framebuffer?: any;
    uniforms?: object;
    discard?: boolean;
    parameters?: object;
    transformFeedback?: any;
};
export declare type TransformBinding = {
    sourceBuffers: Record<string, Buffer>;
    sourceTextures: Record<string, Texture2D>;
    feedbackBuffers?: Record<string, Buffer | {
        buffer: Buffer;
    }>;
    transformFeedback?: TransformFeedback;
    framebuffer?: Framebuffer;
    targetTexture?: Texture2D;
};
/**
 * Takes source and target buffers/textures and sets up the pipeline
 */
export default class Transform {
    /**
     * Check if Transforms are supported (they are not under WebGL1)
     * @todo differentiate writing to buffer vs not
     */
    static isSupported(device: Device | WebGLRenderingContext): boolean;
    readonly device: WebGLDevice;
    readonly gl: WebGL2RenderingContext;
    model: Model | null;
    elementCount: number;
    bufferTransform: BufferTransform;
    textureTransform: TextureTransform;
    elementIDBuffer: Buffer;
    constructor(device: Device | WebGLRenderingContext, props?: TransformProps);
    /** Delete owned resources. */
    destroy(): void;
    /** @deprecated Use destroy*() */
    delete(): void;
    /** Run one transform loop. */
    run(options?: TransformRunOptions): void;
    /** swap resources if a map is provided */
    swap(): void;
    /** Return Buffer object for given varying name. */
    getBuffer(varyingName?: string): Buffer;
    /** Return data either from Buffer or from Texture */
    getData(options?: {
        packed?: boolean;
        varyingName?: string;
    }): any;
    /** Return framebuffer object if rendering to textures */
    getFramebuffer(): Framebuffer | null;
    /** Update some or all buffer/texture bindings. */
    update(props: TransformProps): void;
    _initialize(props: TransformProps): void;
    _updateModelProps(props: TransformProps): TransformProps;
    _buildResourceTransforms(gl: WebGL2RenderingContext, props: TransformProps): void;
    _updateDrawOptions(options: TransformRunOptions): TransformDrawOptions;
}
//# sourceMappingURL=transform.d.ts.map