import _defineProperty from "@babel/runtime/helpers/esm/defineProperty";
import { assert, isObjectEmpty } from '@luma.gl/api';
import { getShaderInfo, getPassthroughFS } from '@luma.gl/shadertools';
import GL from '@luma.gl/constants';
import { WebGLDevice } from '@luma.gl/webgl';
import Model from '../engine/classic-model';
import BufferTransform from './buffer-transform';
import TextureTransform from './texture-transform';
export default class Transform {
  static isSupported(device) {
    try {
      const webglDevice = WebGLDevice.attach(device);
      return webglDevice.isWebGL2;
    } catch {
      return false;
    }
  }

  constructor(device, props = {}) {
    _defineProperty(this, "device", void 0);

    _defineProperty(this, "gl", void 0);

    _defineProperty(this, "model", null);

    _defineProperty(this, "elementCount", 0);

    _defineProperty(this, "bufferTransform", null);

    _defineProperty(this, "textureTransform", null);

    _defineProperty(this, "elementIDBuffer", null);

    this.device = WebGLDevice.attach(device);
    this.gl = this.device.gl2;

    this._initialize(props);

    Object.seal(this);
  }

  destroy() {
    const {
      model,
      bufferTransform,
      textureTransform
    } = this;

    if (model) {
      model.delete();
    }

    if (bufferTransform) {
      bufferTransform.delete();
    }

    if (textureTransform) {
      textureTransform.delete();
    }
  }

  delete() {
    this.destroy();
  }

  run(options) {
    const {
      clearRenderTarget = true
    } = options || {};

    const updatedOpts = this._updateDrawOptions(options);

    if (clearRenderTarget && updatedOpts.framebuffer) {
      updatedOpts.framebuffer.clear({
        color: true
      });
    }

    this.model.transform(updatedOpts);
  }

  swap() {
    let swapped = false;
    const resourceTransforms = [this.bufferTransform, this.textureTransform].filter(Boolean);

    for (const resourceTransform of resourceTransforms) {
      swapped = swapped || resourceTransform.swap();
    }

    assert(swapped, 'Nothing to swap');
  }

  getBuffer(varyingName = null) {
    return this.bufferTransform && this.bufferTransform.getBuffer(varyingName);
  }

  getData(options = {}) {
    const resourceTransforms = [this.bufferTransform, this.textureTransform].filter(Boolean);

    for (const resourceTransform of resourceTransforms) {
      const data = resourceTransform.getData(options);

      if (data) {
        return data;
      }
    }

    return null;
  }

  getFramebuffer() {
    return this.textureTransform && this.textureTransform.getFramebuffer();
  }

  update(props) {
    if ('elementCount' in props) {
      this.model.setVertexCount(props.elementCount);
    }

    const resourceTransforms = [this.bufferTransform, this.textureTransform].filter(Boolean);

    for (const resourceTransform of resourceTransforms) {
      resourceTransform.update(props);
    }
  }

  _initialize(props) {
    const {
      gl
    } = this;

    this._buildResourceTransforms(gl, props);

    props = this._updateModelProps(props);
    this.model = new Model(this.device, { ...props,
      fs: props.fs || getPassthroughFS({
        version: getShaderInfo(props.vs).version
      }),
      id: props.id || 'transform-model',
      drawMode: props.drawMode || GL.POINTS,
      vertexCount: props.elementCount
    });

    if (this.bufferTransform) {
      this.bufferTransform.setupResources({
        model: this.model
      });
    }
  }

  _updateModelProps(props) {
    let updatedProps = { ...props
    };
    const resourceTransforms = [this.bufferTransform, this.textureTransform].filter(Boolean);

    for (const resourceTransform of resourceTransforms) {
      updatedProps = resourceTransform.updateModelProps(updatedProps);
    }

    return updatedProps;
  }

  _buildResourceTransforms(gl, props) {
    if (canCreateBufferTransform(props)) {
      this.bufferTransform = new BufferTransform(gl, props);
    }

    if (canCreateTextureTransform(props)) {
      this.textureTransform = new TextureTransform(gl, props);
    }

    assert(this.bufferTransform || this.textureTransform, 'must provide source/feedback buffers or source/target textures');
  }

  _updateDrawOptions(options) {
    let updatedOpts = { ...options
    };
    const resourceTransforms = [this.bufferTransform, this.textureTransform].filter(Boolean);

    for (const resourceTransform of resourceTransforms) {
      updatedOpts = Object.assign(updatedOpts, resourceTransform.getDrawOptions(updatedOpts));
    }

    return updatedOpts;
  }

}

function canCreateBufferTransform(props) {
  const canCreate = !isObjectEmpty(props.feedbackBuffers) || !isObjectEmpty(props.feedbackMap) || props.varyings && props.varyings.length > 0;
  return Boolean(canCreate);
}

function canCreateTextureTransform(props) {
  const canCreate = !isObjectEmpty(props._sourceTextures) || props._targetTexture || props._targetTextureVarying;
  return Boolean(canCreate);
}
//# sourceMappingURL=transform.js.map